import datetime
import uuid

from typing import List, Optional

import brewman.schemas.ledger as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.Ledger)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["ledger"]),
) -> schemas.Ledger:
    return schemas.Ledger(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        account=None,
        body=[],
    )


@router.get("/{id_}", response_model=schemas.Ledger)
def show_data(
    id_: uuid.UUID,
    request: Request,
    s: Optional[str] = None,
    f: Optional[str] = None,
    user: UserToken = Security(get_user, scopes=["ledger"]),
) -> schemas.Ledger:
    with SessionFuture() as db:
        account = db.execute(select(AccountBase).where(AccountBase.id == id_)).scalar_one()
        start_date = s if s is not None else get_start_date(request.session)
        finish_date = f if f is not None else get_finish_date(request.session)
        body = build_report(account.id, start_date, finish_date, db)
        set_period(start_date, finish_date, request.session)
        return schemas.Ledger(
            startDate=start_date,
            finishDate=finish_date,
            account=schemas.AccountLink(id=account.id, name=account.name),
            body=body,
        )


def build_report(account_id: uuid.UUID, start_date: str, finish_date: str, db: Session) -> List[schemas.LedgerItem]:
    body = []
    opening = opening_balance(account_id, start_date, db)
    body.append(opening)

    query = (
        db.execute(
            select(Voucher)
            .join(Voucher.journals)
            .join(Journal.account)
            .where(
                Voucher.journals.any(Journal.account_id == account_id),
                Voucher.date >= datetime.datetime.strptime(start_date, "%d-%b-%Y"),
                Voucher.date <= datetime.datetime.strptime(finish_date, "%d-%b-%Y"),
                not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            )
            .order_by(Voucher.date, Voucher.last_edit_date)
        )
        .unique()
        .scalars()
        .all()
    )

    for voucher in query:
        debit = 0
        credit = 0
        journal_debit = 0
        name = ""
        for journal in voucher.journals:
            if journal.account_id == account_id:
                journal_debit = journal.debit
                if journal.debit == 1:
                    debit = journal.amount
                    credit = 0
                else:
                    credit = journal.amount
                    debit = 0
        for journal in voucher.journals:
            if journal.debit != journal_debit:
                name += "{0} / ".format(journal.account.name)
        name = name[:-3]
        body.append(
            schemas.LedgerItem(
                id=voucher.id,
                date=voucher.date.strftime("%d-%b-%Y"),
                name=name,
                url=[
                    "/",
                    voucher.voucher_type.name.replace("_", "-").lower(),
                    str(voucher.id),
                ],
                type=voucher.voucher_type.name.replace("_", " ").title(),
                narration=voucher.narration,
                debit=debit,
                credit=credit,
                posted=voucher.posted,
            )
        )
    return body


def opening_balance(account_id: uuid.UUID, start_date: str, db: Session) -> schemas.LedgerItem:
    opening = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .where(
            Voucher.date < datetime.datetime.strptime(start_date, "%d-%b-%Y"),
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            Journal.account_id == account_id,
        )
    ).scalar()
    opening = 0 if opening is None else opening
    if opening < 0:
        credit = opening * -1
        debit = 0
    else:
        debit = opening
        credit = 0
    return schemas.LedgerItem(
        date=start_date,
        id=None,
        name="Opening Balance",
        type="Opening Balance",
        url=[],
        narration="",
        debit=debit,
        credit=credit,
        posted=True,
    )
