import uuid

from datetime import date, datetime
from decimal import Decimal
from typing import List, Optional, Tuple

import brewman.schemas.product_ledger as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy.orm import Session, joinedload
from sqlalchemy.sql.expression import func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.batch import Batch
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.ProductLedger)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["product-ledger"]),
) -> schemas.ProductLedger:
    return schemas.ProductLedger(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        product=None,
        body=[],
    )


@router.get("/{id_}", response_model=schemas.ProductLedger)
def show_data(
    id_: uuid.UUID,
    request: Request,
    s: Optional[str] = None,
    f: Optional[str] = None,
    user: UserToken = Security(get_user, scopes=["product-ledger"]),
) -> schemas.ProductLedger:
    with SessionFuture() as db:
        product = db.execute(select(Product).where(Product.id == id_)).scalar_one()
        start_date = s if s is not None else get_start_date(request.session)
        finish_date = f if f is not None else get_finish_date(request.session)
        body = build_report(
            product.id,
            datetime.strptime(start_date, "%d-%b-%Y"),
            datetime.strptime(finish_date, "%d-%b-%Y"),
            db,
        )
        set_period(start_date, finish_date, request.session)
        return schemas.ProductLedger(
            startDate=start_date,
            finishDate=finish_date,
            product=schemas.ProductLink(id=product.id, name=product.name),
            body=body,
        )


def build_report(
    product_id: uuid.UUID, start_date: date, finish_date: date, db: Session
) -> List[schemas.ProductLedgerItem]:
    running_total_q, running_total_a, opening = opening_balance(product_id, start_date, db)
    body = opening

    query = db.execute(
        select(Voucher, Inventory, Journal, StockKeepingUnit)
        .join(Voucher.journals)
        .join(Voucher.inventories)
        .join(Inventory.batch)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .options(
            joinedload(Journal.account, innerjoin=True),
            joinedload(Journal.cost_centre, innerjoin=True),
        )
        .where(
            StockKeepingUnit.product_id == product_id,
            Journal.cost_centre_id != CostCentre.cost_centre_purchase(),
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
        )
        .order_by(Voucher.date, Voucher.last_edit_date)
    ).all()

    for voucher, inventory, journal, stockKeepingUnit in query:
        journal_debit = journal.debit * -1
        name = (
            journal.cost_centre.name
            if voucher.voucher_type in [VoucherType.ISSUE, VoucherType.CLOSING_STOCK]
            else journal.account.name
        )
        debit_q = inventory.quantity if journal_debit == 1 else None
        debit_a = inventory.amount if journal_debit == 1 else None
        debit_u = stockKeepingUnit.units if journal_debit == 1 else None
        credit_q = inventory.quantity if journal_debit != 1 else None
        credit_a = inventory.amount if journal_debit != 1 else None
        credit_u = stockKeepingUnit.units if journal_debit != 1 else None

        running_total_q += inventory.quantity * journal_debit
        running_total_a += inventory.amount * journal_debit

        body.append(
            schemas.ProductLedgerItem(
                id=voucher.id,
                date=voucher.date,
                name=name,
                url=[
                    "/",
                    voucher.voucher_type.name.replace("_", "-").lower(),
                    str(voucher.id)
                    if voucher.voucher_type != VoucherType.CLOSING_STOCK
                    else voucher.date.strftime("%d-%b-%Y"),
                ],
                type=voucher.voucher_type.name.replace("_", " ").title(),
                narration=voucher.narration,
                posted=voucher.posted or voucher.voucher_type in [VoucherType.ISSUE, VoucherType.CLOSING_STOCK],
                debitQuantity=debit_q,
                debitAmount=debit_a,
                debitUnit=debit_u,
                creditQuantity=credit_q,
                creditAmount=credit_a,
                creditUnit=credit_u,
                runningQuantity=running_total_q,
                runningAmount=running_total_a,
            )
        )

    return body


def opening_balance(
    product_id: uuid.UUID, start_date: date, db: Session
) -> Tuple[Decimal, Decimal, List[schemas.ProductLedgerItem]]:
    row = db.execute(
        select(
            StockKeepingUnit.units,
            func.sum(Inventory.quantity * Journal.debit).label("quantity"),
            func.sum(Inventory.amount * Journal.debit).label("amount"),
        )
        .join(Inventory.batch)
        .join(Batch.sku)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(
            Voucher.id == Inventory.voucher_id,
            Voucher.id == Journal.voucher_id,
            StockKeepingUnit.product_id == product_id,
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
            Voucher.date < start_date,
        )
        .group_by(StockKeepingUnit.units)
    ).all()

    quantity: Decimal = sum(r.quantity for r in row)  # type: ignore[assignment]
    amount: Decimal = sum(r.amount for r in row)  # type: ignore[assignment]

    return (
        quantity,
        amount,
        [
            schemas.ProductLedgerItem(
                id=None,
                date=start_date.strftime("%d-%b-%Y"),
                name="Opening Balance",
                url=[],
                type="Opening Balance",
                narration="",
                posted=True,
                debitQuantity=q,
                debitAmount=a,
                debitUnit=u,
                creditQuantity=0,
                creditAmount=0,
                creditUnit="",
                runningQuantity=quantity,
                runningAmount=amount,
            )
            for u, q, a in row
        ],
    )
